import ipaddress
import logging
import os
import pwd
import grp
import re
import socket
import subprocess
import time
from logging.handlers import RotatingFileHandler

import redis

class HAManagerException(Exception):
    """Exception to short-circuit the processing loop and continue with the next iteration."""

class HAManager:
    """
    The HA Manager is a set of logic designed to be run as a service,
    which manages routes on a DreamPress podman host in order to keep
    a host full of websites highly available. In general terms, this is
    done using a centralized redis datastore and checkins from the active
    and passive hosts. The sites are failved over if a host doesn't check
    in for some period of time.

    The routes that this service installs on the local host are then
    used by FRR's `redistribute kernel` bgp function to advertise each
    site's floating IP.
    """

    MAX_HEARTBEAT_AGE = 300
    SOCKET_TIMEOUT = 0.2

    def __init__(self):
        """
        Initialize the class by setting up logging, connecting to Redis,
        and initializing other necessary attributes.
        """
        self.logger = self.setup_logging()
        self.hostname = socket.gethostname()
        self.hostgroup = "__UNSET__"
        self.hostgroup_data = {}
        self.current_routes = {}
        self.desired_routes = {}

        # recent_redis_errors keeps track of how many redis errors have been encountered
        # in the last 10 loops, and is used to determine whether or not it's safe to
        # take actions like failing over and whether to log warnings or errors.
        # The range is 0-10, and it starts in the middle for safety, since when the service
        # starts we don't actually know the status of redis connectivity/health
        self.recent_redis_errors = 5

        try:
            self.redis = self.setup_redis_connection()
        except redis.exceptions.RedisError as e:
            self.logger.error(f"Redis exception: {e}")

    def setup_redis_connection(self):
        """
        Set up a connection to the Redis server.
        """
        sentinel_hosts = self.get_sentinel_hosts()
        sentinel_service_name = os.environ.get('REDIS_SENTINEL_SERVICE')
        sentinel_password = os.environ.get('REDIS_SENTINEL_PASSWORD')
        redis_password = os.environ.get('REDIS_PASSWORD')

        sentinel = redis.Sentinel(
                sentinel_hosts,
                sentinel_kwargs={'password': sentinel_password})

        return sentinel.master_for(
                sentinel_service_name,
                socket_timeout=self.SOCKET_TIMEOUT,
                username=self.hostname,
                password=redis_password)

    def get_sentinel_hosts(self):
        """
        Retrieve Redis sentinel host information from environment variables.
        """
        # REDIS_SENTINEL_HOSTS env var should be a comma separated list of sentinel hosts
        sentinel_hosts_env = os.environ.get('REDIS_SENTINEL_HOSTS')
        sentinel_port = os.environ.get('REDIS_SENTINEL_PORT')
        sentinel_hosts = []

        if not sentinel_hosts_env:
            raise ValueError("No Redis sentinel hosts found in environment variable 'REDIS_SENTINEL_HOSTS'.")
        if not sentinel_port:
            raise ValueError("No Redis sentinel port found in environment variable 'REDIS_SENTINEL_PORT'.")

        # Split the environment variable by comma to get the list of hostnames
        hosts = sentinel_hosts_env.split(',')
        # Append each host to the sentinel_hosts list
        for host in hosts:
            sentinel_hosts.append((host.strip(), sentinel_port))  # .strip() removes any leading/trailing whitespace

        if not sentinel_hosts:
            raise ValueError(f"No hosts parsed from the 'REDIS_SENTINEL_HOSTS' env var: {sentinel_hosts_env}")

        return sentinel_hosts

    def setup_logging(self):
        """
        Python logger configuration
        """
        logger = logging.getLogger('HaManagerLogger')
        logger.setLevel(logging.INFO)
        handler = RotatingFileHandler(
                os.environ.get('LOG_PATH') + "/hamanager.log",
                maxBytes=1000000,
                backupCount=5)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return logger

    def write_env_file(self):
        """
        Writes an environment variable file for consumption by other services
        """
        # Unix epoch time in seconds
        epoch_time = int(time.time())
        env_file = "/root/.hamanager_data"
        i_am_current_primary = self.i_am_current_primary()

        # Construct the content to write to the file
        content = "# HAManager data -- do not delete. This is used as input for other services.\n"
        content += f"HA_TIMESTAMP={epoch_time}\n"
        content += f"HA_I_AM_PRIMARY={str(i_am_current_primary).upper()}\n"
        uid = pwd.getpwnam("root").pw_uid
        gid = grp.getgrnam("root").gr_gid

        try:
            # Write the content to the file
            with open(env_file, "w", encoding="utf-8") as file:
                file.write(content)
            os.chown(env_file, uid, gid)
        except Exception as e: # pylint: disable=broad-exception-caught
            self.logger.exception(e)
            self.logger.error(f"Error writing data to {env_file}. This will cause problems!")

    def get_current_routes(self):
        """
        Gets all the current host routes on this machine, and
        returns them as a dict.
        """
        command = ["ip", "route"]
        try:
            output = subprocess.check_output(command, text=True)
        except subprocess.CalledProcessError as e:
            print(f"Failed to execute command: {e}")
            return {}

        current_routes = {}
        for line in output.splitlines():
            match = re.match(r"(\d+\.\d+\.\d+\.\d+) via (\d+\.\d+\.\d+\.\d+) dev private", line)
            if match:
                ip, nexthop = match.groups()
                current_routes[ip] = nexthop

        return current_routes

    def is_host_stale(self, host_to_check):
        """
        Takes a hostname and checks to see if the host has failed
        to check in with redis for the last "MAX_HEARTBEAT_AGE" seconds
        """
        heartbeat = self.redis.get("heartbeat." + host_to_check)
        if heartbeat is not None:
            # Convert the heartbeat to an integer
            last_heartbeat = int(heartbeat)
            # Get the current time
            current_time = int(time.time())
            # Calculate the difference
            time_difference = current_time - last_heartbeat

            # Compare the time difference with the threshold.
            # If the difference is less than the threshold, the
            # host has checked in with redis recently and is not stale
            if time_difference <= self.MAX_HEARTBEAT_AGE:
                return False  # Host isn't stale

            # The host passed in is stale, since it's been longer
            # than the threshold since it last checked in with redis
            self.logger.warning(f"Heartbeat for '{host_to_check}' is stale (exceeds {self.MAX_HEARTBEAT_AGE} seconds).")
            return True

        # If for some reason there is no heartbeat at all for the host:
        self.logger.error(f"No heartbeat found for host '{host_to_check}'")
        return True  # No heartbeat found for host

    def fetch_and_decode_hostgroup_data(self):
        """
        Fetches and returns decoded hostgroup data from redis.
        If the data cannot be retrieved, raise an error
        """
        data = self.redis.hgetall(self.hostgroup)
        if data:
            return {key.decode('utf-8'): value.decode('utf-8') for key, value in data.items()}
        raise HAManagerException(f"Failed to fetch hostgroup data for hostgroup '{self.hostgroup}'")

    def i_am_current_primary(self):
        """
        Returns true if this host is the current_primary
        """
        return self.hostgroup_data.get("config.current_primary") == self.hostname

    def i_am_defined_primary(self):
        """
        Returns true if this host is the defined_primary
        """
        return self.hostgroup_data.get("config.defined_primary") == self.hostname

    def valid_ip(self, ip, service_name):
        """
        Accepts an IP and DP service name, and returns a boolean
        representing whether ip is a valid IP address. Also logs
        an error if the IP is invalid.
        """
        try:
            ipaddress.ip_address(ip)
        except ValueError:
            self.logger.warning(f"IP address '{ip}' is invalid for {service_name}.")
            return False
        return True

    def get_desired_routes(self):
        """
        If this machine is current_primary, figure out all the routes that need
        to be installed, and return them. If we are not current_primary, we don't
        want any routes to be installed, so return an empty dict.
        """
        # If this machine is primary, add the IPs
        desired_routes = {}
        if self.i_am_current_primary():
            for key, value in self.hostgroup_data.items():
                # Check if the key is related to service IP configuration
                # example key: service.dp-test-000000.primary_ip
                if key.startswith("service.") and key.endswith(".floating_ip"):
                    service_name = key.split(".")[1]
                    # Make sure we get the right next-hop IP based on whether we're
                    # the defined primary or standby
                    if self.i_am_defined_primary():
                        private_ip = self.hostgroup_data.get(f"service.{service_name}.primary_ip")
                    else:
                        private_ip = self.hostgroup_data.get(f"service.{service_name}.standby_ip")

                    # Ensure we are working with valid IPs
                    if not self.valid_ip(private_ip, service_name):
                        continue
                    if not self.valid_ip(value, service_name):
                        continue

                    # Add the route to the desired_routes dict if it's valid
                    desired_routes[value] = private_ip
        return desired_routes

    def standby_checks(self):
        """
        Checks for 'stale primary' and 'fail back' scenarios
        and updates redis if necessary.
        """
        if self.i_am_current_primary():
            # we're not the current standby, don't need to do anything
            return

        current_primary = self.hostgroup_data.get("config.current_primary")
        defined_primary = self.hostgroup_data.get("config.defined_primary")
        revert_allowed = int(self.hostgroup_data.get("config.revert_allowed", 0))

        if self.is_host_stale(defined_primary) and self.recent_redis_errors == 0:
            # Current primary is stale. We need to take over. Update current_primary
            # to this machine, so that the floating IP will be advertised on the next loop.
            self.redis.hset(self.hostgroup, "config.current_primary", self.hostname)
            self.logger.warning(f"Updated 'current_primary' for hostgroup '{self.hostgroup}' to '{self.hostname}'")

        # Additional check: If a failover has already happened and we're the defined_primary,
        # we are probably in a fail-back scenario. If a revert is allowed we need to
        # set ourselves as current primary and reset revert_allowed.
        if (current_primary != defined_primary and
                defined_primary == self.hostname and
                revert_allowed == 1 and
                self.recent_redis_errors == 0):
            self.redis.hset(self.hostgroup, "config.current_primary", self.hostname)
            self.redis.hset(self.hostgroup, "config.revert_allowed", 0)
            self.logger.warning(f"Updated 'current_primary' for hostgroup '{self.hostgroup}' to '{self.hostname}'")
            self.logger.info(f"Updated 'revert_allowed' for hostgroup '{self.hostgroup}' to '0'")

    def update_routes(self):
        """
        Add missing routes and remove unnecesary routes. This is done by comparing
        the desired routes with the current routes.
        """
        # Add missing routes
        for ip, nexthop in self.desired_routes.items():
            if ip not in self.current_routes or self.current_routes[ip] != nexthop:
                try:
                    subprocess.run(["ip", "route", "add", ip, "via", nexthop], check=True)
                except Exception as e: # pylint: disable=broad-except
                    self.logger.exception(e)
                    self.logger.error(f"Error adding route {ip} via {nexthop}")
                self.logger.info(f"Added route: {ip} via {nexthop}")

        # Remove any extra routes
        for ip, nexthop in self.current_routes.items():
            if ip not in self.desired_routes or self.desired_routes[ip] != nexthop:
                try:
                    subprocess.run(["ip", "route", "del", ip, "via", nexthop], check=True)
                except Exception as e: # pylint: disable=broad-except
                    self.logger.exception(e)
                    self.logger.error(f"Error removing route {ip} via {nexthop}")
                self.logger.info(f"Removed route: {ip} via {nexthop}")

    def send_heartbeat(self):
        """
        Update this machine's heartbeat timestamp in redis,
        so that other machines can see that this one is up
        and running.
        """
        timestamp = int(time.time())
        self.redis.set("heartbeat." + self.hostname, timestamp)

    def ensure_hostgroup_is_set(self):
        """
        If the hostgroup has not already been set, this will fetch the hostgroup from redis based on our hostname.
        If the hostgroup has already been set, this will do nothing
        """
        if self.hostgroup == "__UNSET__":
            # Attempt to retrieve the hostgroup using our hostname
            hostgroup = self.redis.get(f"{self.hostname}.hostgroup")
            if hostgroup:
                self.hostgroup = hostgroup.decode('utf-8')
            else:
                raise HAManagerException(f"Error determining hostgroup membership for local host: {self.hostname}")

    def redis_failure(self, err):
        """
        This function is called when a redis exception is caught, and
        logs either a warning or an error depending on how many recent
        redis errors have occurred.
        """
        if self.recent_redis_errors < 10:
            self.logger.warning(f"Redis exception: {err}")
            self.recent_redis_errors += 1

        if self.recent_redis_errors == 10:
            self.logger.error(f"Redis exception: {err}")

    def route_check(self):
        """
        This is the logic that gets executed each second
        for checking and updating routes on the host
        """
        self.ensure_hostgroup_is_set()
        self.hostgroup_data = self.fetch_and_decode_hostgroup_data()
        self.current_routes = self.get_current_routes()
        self.send_heartbeat()
        self.desired_routes = self.get_desired_routes()
        self.standby_checks()
        self.update_routes()
        self.write_env_file()

    def run(self):
        """
        This function contains the main "while True" loop of the service
        """
        self.logger.info("Podman HA Manager starting up.")
        start_time = time.time()
        total_loop_time = 0
        loop_count = 0

        while True:
            loop_start_time = time.time()

            try:
                self.route_check()
                if self.recent_redis_errors > 0:
                    self.recent_redis_errors -= 1
            except redis.exceptions.RedisError as e:
                self.redis_failure(e)
            except HAManagerException as e:
                self.logger.error(f"Hamanager exception: {e}")
            except Exception as e:
                self.logger.exception(e)
                raise e  # Service fails

            # Determine how long is each loop taking
            loop_duration = time.time() - loop_start_time
            total_loop_time += loop_duration
            loop_count += 1

            # Check if a minute has passed
            if time.time() - start_time >= 60:  # 60 seconds = 1 minute
                average_time = total_loop_time / loop_count if loop_count else 0
                self.logger.info(f"Average loop time over the last minute: {average_time:.2f} seconds")
                # Reset counters
                start_time = time.time()
                total_loop_time = 0
                loop_count = 0

            # Sleep for 1s after logic and any exceptions
            time.sleep(1)

if __name__ == "__main__":
    manager = HAManager()
    manager.run()
